/*
 *  Part of the foobar2000 Matroska plugin
 *
 *  Copyright (C) Jory 'jcsston' Stone (jcsston at toughguy net) - 2003-2004
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 *  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*!
    \file matroska_parser.cpp
		\version $Id$
    \brief An audio slated Matroska Reader+Parser
		\author Jory Stone     <jcsston @ toughguy.net>
*/

#include "matroska_parser.h"

#include <string>
using std::string;

using namespace LIBEBML_NAMESPACE;	
using namespace LIBMATROSKA_NAMESPACE;

namespace Matroska {

bool starts_with(const string &s, const char *start) {
  return strncmp(s.c_str(), start, strlen(start)) == 0;
}

MatroskaAttachment::MatroskaAttachment()
{
	FileName = L"";
	MimeType = "";
	Description = L"";
	SourceFilename = L"";
	SourceStartPos = 0;
	SourceDataLength = 0;
}

MatroskaEditionInfo::MatroskaEditionInfo() {
	editionUID = 0;
}

MatroskaSimpleTag::MatroskaSimpleTag()
{
	name = L"";
	value = L"";
	language = "und";
	defaultFlag = 1;
	hidden = false;
	removalPending = false;
};

MatroskaTagInfo::MatroskaTagInfo()
{
	targetTrackUID = 0;
	targetEditionUID = 0;
	targetChapterUID = 0;
	targetAttachmentUID = 0;
	targetTypeValue = 0;
};

void MatroskaTagInfo::SetTagValue(const char *name, const char *value, int index)
{
	for (size_t s = 0; s < tags.size(); s++)
	{
		MatroskaSimpleTag &currentSimpleTag = tags.at(s);
		if (strcmpi(currentSimpleTag.name.GetUTF8().c_str(), name) == 0)			
		{
			if(index == 0)
			{
				currentSimpleTag.value.SetUTF8(value);
				currentSimpleTag.removalPending = false;
				return;
			}
			index--;
		}
	}

	// If we are here then we didn't find this tag in the vector already
	MatroskaSimpleTag newSimpleTag;
	newSimpleTag.name.SetUTF8(name);
	newSimpleTag.value.SetUTF8(value);	
	newSimpleTag.removalPending = false;
	tags.push_back(newSimpleTag);
};

void MatroskaTagInfo::RemoveMarkedTags()
{
	for (int i = tags.size()-1; i >= 0; i--)
	{
		MatroskaSimpleTag &simpleTag = tags.at(i);
		if(simpleTag.removalPending == true)
		{
			tags.erase(tags.begin() + i);
		}
	}
}

void MatroskaTagInfo::MarkAllAsRemovalPending()
{
	for (int i = tags.size()-1; i >= 0; i--)
	{
		MatroskaSimpleTag &simpleTag = tags.at(i);
		simpleTag.removalPending = true;
	}
}

MatroskaChapterDisplayInfo::MatroskaChapterDisplayInfo()  {
	string = L"";
};

MatroskaChapterInfo::MatroskaChapterInfo() {
	chapterUID = 0;
	timeStart = 0;
	timeEnd = 0;
}

MatroskaAudioParser::MatroskaAudioParser(service_ptr_t<file> input, abort_callback & p_abort) 
	: m_IOCallback(input, p_abort),
		m_InputStream(m_IOCallback)
{
	m_TimecodeScale = TIMECODE_SCALE;
	m_FileDate = 0;
	m_Duration = 0;
	m_CurrentTimecode = 0;
	m_ElementLevel0 = NULL;
	//UpperElementLevel = 0;
	m_FileSize = input->get_size(p_abort);
	m_TagPos = 0;
	m_TagSize = 0;
	m_TagScanRange = 1024 * 64;
	m_CurrentTrackNo = 0;
	m_SegmentFilename = L"";
};

MatroskaAudioParser::~MatroskaAudioParser() {
	if (m_ElementLevel0 != NULL)
		delete m_ElementLevel0;
};

int MatroskaAudioParser::Parse(bool bInfoOnly, bool bBreakAtClusters) 
{
	try {
		int UpperElementLevel = 0;
		bool bAllowDummy = false;
		// Elements for different levels
		EbmlElement *ElementLevel1 = NULL;
		EbmlElement *ElementLevel2 = NULL;
		EbmlElement *ElementLevel3 = NULL;
		EbmlElement *ElementLevel4 = NULL;
		EbmlElement *ElementLevel5 = NULL;		

		// Be sure we are at the beginning of the file
		m_IOCallback.setFilePointer(0);
		// Find the EbmlHead element. Must be the first one.
		m_ElementLevel0 = m_InputStream.FindNextID(EbmlHead::ClassInfos, 0xFFFFFFFFFFFFFFFFL);
		if (m_ElementLevel0 == NULL) {
			return 1;
		}
		//We must have found the EBML head :)
		m_ElementLevel0->SkipData(m_InputStream, m_ElementLevel0->Generic().Context);
		delete m_ElementLevel0;

		// Next element must be a segment
		m_ElementLevel0 = m_InputStream.FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFFFFFFFFFL);
		if (m_ElementLevel0 == NULL) {
			//No segment/level 0 element found.
			return 1;
		}
		if (!(EbmlId(*m_ElementLevel0) == KaxSegment::ClassInfos.GlobalId)) {
			delete m_ElementLevel0;
			m_ElementLevel0 = NULL;
			return 1;
		}

		UpperElementLevel = 0;
		// We've got our segment, so let's find the tracks
		ElementLevel1 = m_InputStream.FindNextElement(m_ElementLevel0->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, true, 1);
		while (ElementLevel1 != NULL) {
			if (UpperElementLevel > 0) {
				break;
			}
			if (UpperElementLevel < 0) {
				UpperElementLevel = 0;
			}
			if (EbmlId(*ElementLevel1) == KaxSeekHead::ClassInfos.GlobalId) {
				if (m_IOCallback.seekable()) {
					Parse_MetaSeek(ElementLevel1, bInfoOnly);
					if (m_TagPos == 0) {
						// Search for them at the end of the file
						if (m_TagScanRange > 0)
						{
							TIMER;
							m_IOCallback.setFilePointer(m_FileSize - m_TagScanRange);
							uint64 init_pos = m_IOCallback.getFilePointer();
							/*
								BM Search
							*/
							binary* buf = new binary[1024*64];
							binary* pat = new binary[3]; pat[0] = 0x54; pat[1] = 0xc3; pat[2] = 0x67;
							uint64 s_pos = m_IOCallback.getFilePointer();
							m_IOCallback.read(buf, m_TagScanRange);
							MatroskaSearch search(buf, pat);
							int pos = search.Match();
							if (pos != -1) {
								do {
									m_IOCallback.setFilePointer(s_pos+pos+3);
									uint64 startPos = m_IOCallback.getFilePointer();
									m_IOCallback.setFilePointer(-4, seek_current);
									EbmlElement *levelUnknown = m_InputStream.FindNextID(KaxTags::ClassInfos, 0xFFFFFFFFFFFFFFFFL);
									if ((levelUnknown != NULL) 
										&& (m_FileSize >= startPos + levelUnknown->GetSize()) 
										&& (EbmlId(*levelUnknown) == KaxTags::ClassInfos.GlobalId))
									{
										Parse_Tags(static_cast<KaxTags *>(levelUnknown));
										_TIMER("MatroskaAudioParser::Parse(BM)");
										break;
									}
									delete levelUnknown;
									levelUnknown = NULL;
									m_IOCallback.setFilePointer(s_pos);
									pos = search.Match(pos+1);
								} while (pos != -1);
								/*
									~BM Search
								*/
							} else {
								m_IOCallback.setFilePointer(init_pos);
								binary Buffer[4];
								while (m_IOCallback.read(Buffer, 3) >= 3)
								{//0x18
									if ((Buffer[0] == 0x54) && (Buffer[1] == 0xc3) && (Buffer[2] == 0x67))
									{
										uint64 startPos = m_IOCallback.getFilePointer();

										//seek back 3 bytes, so libmatroska can find the Tags element Ebml ID
										m_IOCallback.setFilePointer(-4, seek_current);

										EbmlElement *levelUnknown = m_InputStream.FindNextID(KaxTags::ClassInfos, 0xFFFFFFFFFFFFFFFFL);
										if ((levelUnknown != NULL) 
											&& (m_FileSize >= startPos + levelUnknown->GetSize()) 
											&& (EbmlId(*levelUnknown) == KaxTags::ClassInfos.GlobalId))
										{
											Parse_Tags(static_cast<KaxTags *>(levelUnknown));
											delete levelUnknown;
											levelUnknown = NULL;
											break;
										}
										delete levelUnknown;
										levelUnknown = NULL;

										//Restore the file pos
										m_IOCallback.setFilePointer(startPos);
									}
									//seek back 2 bytes
									m_IOCallback.setFilePointer(-2, seek_current);
								}
							}
						} else {
							//m_TagPos = m_FileSize;
						}
					}
				}
			}else if (EbmlId(*ElementLevel1) == KaxInfo::ClassInfos.GlobalId) {
				// General info about this Matroska file
				ElementLevel2 = m_InputStream.FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, bAllowDummy);
				while (ElementLevel2 != NULL) {
					if (UpperElementLevel > 0) {
						break;
					}
					if (UpperElementLevel < 0) {
						UpperElementLevel = 0;
					}

					if (EbmlId(*ElementLevel2) == KaxTimecodeScale::ClassInfos.GlobalId) {
						KaxTimecodeScale &TimeScale = *static_cast<KaxTimecodeScale *>(ElementLevel2);
						TimeScale.ReadData(m_InputStream.I_O());

						//matroskaGlobalTrack->SetTimecodeScale(uint64(TimeScale));
						m_TimecodeScale = uint64(TimeScale);
					} else if (EbmlId(*ElementLevel2) == KaxDuration::ClassInfos.GlobalId) {
						KaxDuration &duration = *static_cast<KaxDuration *>(ElementLevel2);
						duration.ReadData(m_InputStream.I_O());

						// it's in milliseconds?
						m_Duration = double(duration) * (double)(int64)m_TimecodeScale;
						//matroskaGlobalTrack->SetDuration(float(duration) * int64(m_TimecodeScale) / 1000000000.0);

					} else if (EbmlId(*ElementLevel2) == KaxDateUTC::ClassInfos.GlobalId) {
						KaxDateUTC & DateUTC = *static_cast<KaxDateUTC *>(ElementLevel2);
						DateUTC.ReadData(m_InputStream.I_O());
						
						m_FileDate = DateUTC.GetEpochDate();

					} else if (EbmlId(*ElementLevel2) == KaxSegmentFilename::ClassInfos.GlobalId) {
						KaxSegmentFilename &tag_SegmentFilename = *static_cast<KaxSegmentFilename *>(ElementLevel2);
						tag_SegmentFilename.ReadData(m_InputStream.I_O());

						m_SegmentFilename = *static_cast<EbmlUnicodeString *>(&tag_SegmentFilename);

					} else if (EbmlId(*ElementLevel2) == KaxMuxingApp::ClassInfos.GlobalId)	{
						KaxMuxingApp &tag_MuxingApp = *static_cast<KaxMuxingApp *>(ElementLevel2);
						tag_MuxingApp.ReadData(m_InputStream.I_O());

						m_MuxingApp = *static_cast<EbmlUnicodeString *>(&tag_MuxingApp);

					} else if (EbmlId(*ElementLevel2) == KaxWritingApp::ClassInfos.GlobalId) {
						KaxWritingApp &tag_WritingApp = *static_cast<KaxWritingApp *>(ElementLevel2);
						tag_WritingApp.ReadData(m_InputStream.I_O());
						
						m_WritingApp = *static_cast<EbmlUnicodeString *>(&tag_WritingApp);

					} else if (EbmlId(*ElementLevel2) == KaxTitle::ClassInfos.GlobalId) {
						KaxTitle &Title = *static_cast<KaxTitle*>(ElementLevel2);
						Title.ReadData(m_InputStream.I_O());
						m_FileTitle = UTFstring(Title).c_str();
					}

					if (UpperElementLevel > 0) {	// we're coming from ElementLevel3
						UpperElementLevel--;
						delete ElementLevel2;
						ElementLevel2 = ElementLevel3;
						if (UpperElementLevel > 0)
							break;
					} else {
						ElementLevel2->SkipData(m_InputStream, ElementLevel2->Generic().Context);
						delete ElementLevel2;
						ElementLevel2 = m_InputStream.FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, bAllowDummy);
					}
				}
			} else if (EbmlId(*ElementLevel1) == KaxCluster::ClassInfos.GlobalId) {
				
				if (bBreakAtClusters) {
					m_IOCallback.setFilePointer(ElementLevel1->GetElementPosition());
					delete ElementLevel1;
					ElementLevel1 = NULL;
					break;
				}

			} else if (EbmlId(*ElementLevel1) == KaxAttachments::ClassInfos.GlobalId) {
				// Yep, we've found our KaxAttachment element. Now find all attached files
				// contained in this segment.
#if 1
				ElementLevel2 = m_InputStream.FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, true, 1);
				while (ElementLevel2 != NULL) {
					if (UpperElementLevel > 0) {
						break;
					}
					if (UpperElementLevel < 0) {
						UpperElementLevel = 0;
					}
					if (EbmlId(*ElementLevel2) == KaxAttached::ClassInfos.GlobalId) {
						// We actually found a attached file entry :D
						MatroskaAttachment newAttachment;

						ElementLevel3 = m_InputStream.FindNextElement(ElementLevel2->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, true, 1);
						while (ElementLevel3 != NULL) {
							if (UpperElementLevel > 0) {
								break;
							}
							if (UpperElementLevel < 0) {
								UpperElementLevel = 0;
							}

							// Now evaluate the data belonging to this track
							if (EbmlId(*ElementLevel3) == KaxFileName::ClassInfos.GlobalId) {
								KaxFileName &attached_filename = *static_cast<KaxFileName *>(ElementLevel3);
								attached_filename.ReadData(m_InputStream.I_O());
								newAttachment.FileName = UTFstring(attached_filename);

							} else if (EbmlId(*ElementLevel3) == KaxMimeType::ClassInfos.GlobalId) {
								KaxMimeType &attached_mime_type = *static_cast<KaxMimeType *>(ElementLevel3);
								attached_mime_type.ReadData(m_InputStream.I_O());
								newAttachment.MimeType = std::string(attached_mime_type);

							} else if (EbmlId(*ElementLevel3) == KaxFileDescription::ClassInfos.GlobalId) {
								KaxFileDescription &attached_description = *static_cast<KaxFileDescription *>(ElementLevel3);
								attached_description.ReadData(m_InputStream.I_O());
								newAttachment.Description = UTFstring(attached_description);

							} else if (EbmlId(*ElementLevel3) == KaxFileData::ClassInfos.GlobalId) {
								KaxFileData &attached_data = *static_cast<KaxFileData *>(ElementLevel3);

								//We don't what to read the data into memory because it could be very large
								//attached_data.ReadData(m_InputStream.I_O());

								//Instead we store the Matroska filename, the start of the data and the length, so we can read it
								//later at the users request. IMHO This will save a lot of memory
								newAttachment.SourceFilename = GetSegmentFileName();
								newAttachment.SourceStartPos = attached_data.GetElementPosition() + attached_data.HeadSize();
								newAttachment.SourceDataLength = attached_data.GetSize();
							}

							if (UpperElementLevel > 0) {	// we're coming from ElementLevel4
								UpperElementLevel--;
								delete ElementLevel3;
								ElementLevel3 = ElementLevel4;
								if (UpperElementLevel > 0)
									break;
							} else {
								ElementLevel3->SkipData(m_InputStream, ElementLevel3->Generic().Context);
								delete ElementLevel3;
								ElementLevel3 = m_InputStream.FindNextElement(ElementLevel2->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, true, 1);
							}					
						} // while (ElementLevel3 != NULL)
						m_AttachmentList.push_back(newAttachment);
					}

					if (UpperElementLevel > 0) {	// we're coming from ElementLevel3
						UpperElementLevel--;
						delete ElementLevel2;
						ElementLevel2 = ElementLevel3;
						if (UpperElementLevel > 0)
							break;
					} else {
						ElementLevel2->SkipData(m_InputStream, ElementLevel2->Generic().Context);
						delete ElementLevel2;
						ElementLevel2 = m_InputStream.FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, true, 1);
					}
				} // while (ElementLevel2 != NULL)
#endif
			}

			
			if (UpperElementLevel > 0) {		// we're coming from ElementLevel2
				UpperElementLevel--;
				delete ElementLevel1;
				ElementLevel1 = ElementLevel2;
				if (UpperElementLevel > 0)
					break;
			} else {
				ElementLevel1->SkipData(m_InputStream, ElementLevel1->Generic().Context);
				delete ElementLevel1;
				ElementLevel1 = NULL;

				ElementLevel1 = m_InputStream.FindNextElement(m_ElementLevel0->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, true, 1);
			}
		} // while (ElementLevel1 != NULL)		
	
	} catch (std::exception &) {
		return 1;

	} catch (...) {
		return 1;
	}

	CountClusters();
	return 0;
};

int meta_get_num(const file_info &info, const char* name, int idx)
{
	assert(pfc::is_valid_utf8(name));
	int n, m = min(info.meta_get_count(), idx);
	int rv = 0;
	for(n=0; n<m; n++)
	{
		if (!stricmp_utf8(name, info.meta_enum_name(n)))
			rv++;
	}
	return rv;
}

#define IS_ELEMENT_ID(__x__) (Element->Generic().GlobalId == __x__::ClassInfos.GlobalId)

void MatroskaAudioParser::Parse_Chapter_Atom(KaxChapterAtom *ChapterAtom)
{
	int i, j;
	EbmlElement *Element = NULL;
	MatroskaChapterInfo newChapter;

	NOTE("New chapter");

	for (i = 0; i < ChapterAtom->ListSize(); i++)
	{	
		Element = (*ChapterAtom)[i];
				
		if (IS_ELEMENT_ID(KaxChapterUID))
		{
			newChapter.chapterUID = uint64(*static_cast<EbmlUInteger *>(Element));
			NOTE1("- UID : %I64d", newChapter.chapterUID);
		}
		else if(IS_ELEMENT_ID(KaxChapterTimeStart))
		{							
			newChapter.timeStart = uint64(*static_cast<EbmlUInteger *>(Element)); // it's in ns
			NOTE1("- TimeStart : %I64d", newChapter.timeStart);
		}
		else if(IS_ELEMENT_ID(KaxChapterTimeEnd))
		{
			newChapter.timeEnd = uint64(*static_cast<EbmlUInteger *>(Element)); // it's in ns
			NOTE1("- TimeEnd : %I64d", newChapter.timeEnd);
		}
		else if(IS_ELEMENT_ID(KaxChapterTrack))
		{
			KaxChapterTrack *ChapterTrack = (KaxChapterTrack *)Element;
			
			for (j = 0; j < ChapterTrack->ListSize(); j++)
			{
				Element = (*ChapterTrack)[j];
				if(IS_ELEMENT_ID(KaxChapterTrackNumber))
				{
					uint64 chapTrackNo = uint64(*static_cast<EbmlUInteger *>(Element));
					newChapter.tracks.push_back(chapTrackNo);
					NOTE1("- TrackNumber : %I64d", chapTrackNo);
				}
				else if(IS_ELEMENT_ID(KaxChapterAtom))
				{									
					// Ignore sub-chapter
					NOTE("ignore sub-chapter");
					//Parse_Chapter_Atom((KaxChapterAtom *)Element);
				}
			}
		}
		else if(IS_ELEMENT_ID(KaxChapterDisplay))
		{
			// A new chapter display string+lang+country
			MatroskaChapterDisplayInfo newChapterDisplay;							
			KaxChapterDisplay *ChapterDisplay = (KaxChapterDisplay *)Element;
			
			for (j = 0; j < ChapterDisplay->ListSize(); j++) {
				Element = (*ChapterDisplay)[j];
				if(IS_ELEMENT_ID(KaxChapterString))
				{
					newChapterDisplay.string = UTFstring(*static_cast <EbmlUnicodeString *>(Element)).c_str();
					NOTE1("- String : %s", newChapterDisplay.string.GetUTF8().c_str());
				}
				else if(IS_ELEMENT_ID(KaxChapterAtom))
				{									
					// Ignore sub-chapter
					NOTE("ignore sub-chapter");
					//Parse_Chapter_Atom((KaxChapterAtom *)Element);
				}
			}
			// A emtpy string in a chapter display string is usless
			if (newChapterDisplay.string.length() > 0)
				newChapter.display.push_back(newChapterDisplay);
		}
		else if(IS_ELEMENT_ID(KaxChapterAtom))
		{							
			// Ignore sub-chapter
			NOTE("ignore sub-chapter");
			//Parse_Chapter_Atom((KaxChapterAtom *)Element);
		}
	}
	if ((newChapter.chapterUID != 0) && !FindChapterUID(newChapter.chapterUID))
		m_Chapters.push_back(newChapter);
}

void MatroskaAudioParser::Parse_Chapters(KaxChapters *chaptersElement)
{
	int i, j;	
	EbmlElement *Element = NULL;	
	int UpperEltFound = 0;

	NOTE("New edition");

	if (chaptersElement == NULL)
		return;

	chaptersElement->Read(m_InputStream, KaxChapters::ClassInfos.Context,
		UpperEltFound, Element, true);

	for (i = 0; i < chaptersElement->ListSize(); i++)
	{
		Element = (*chaptersElement)[i];
		if(IS_ELEMENT_ID(KaxEditionEntry))
		{
			MatroskaEditionInfo newEdition;
			KaxEditionEntry *edition = (KaxEditionEntry *)Element;
			for (j = 0; j < edition->ListSize(); j++)
			{
				Element = (*edition)[j];
				if(IS_ELEMENT_ID(KaxEditionUID))
				{
					// A new edition :)
					newEdition.editionUID = uint64(*static_cast<EbmlUInteger *>(Element));
					NOTE1("- UID : %I64d", newEdition.editionUID);
				}
				else if(IS_ELEMENT_ID(KaxChapterAtom))
				{
					// A new chapter :)
					Parse_Chapter_Atom((KaxChapterAtom *)Element);
				}
			}
			if ((newEdition.editionUID != 0) && !FindEditionUID(newEdition.editionUID))
				m_Editions.push_back(newEdition);
		}
	}
	FixChapterEndTimes();
}

void MatroskaAudioParser::Parse_Tags(KaxTags *tagsElement)
{
	int i, j, k;
	EbmlElement *Element = NULL;
	int UpperEltFound = 0;

	if (tagsElement == NULL)
		return;

	m_TagPos = tagsElement->GetElementPosition();
	m_TagSize = tagsElement->GetSize();

	tagsElement->Read(m_InputStream, KaxTags::ClassInfos.Context, UpperEltFound, Element, true);

	for (i = 0; i < tagsElement->ListSize(); i++)
	{
		Element = (*tagsElement)[i];
		if(IS_ELEMENT_ID(KaxTag))
		{
			MatroskaTagInfo newTag;
			newTag.targetTypeValue = 50;
			KaxTag *tagElement = (KaxTag*)Element;
			NOTE("New Tag");
			for (j = 0; j < tagElement->ListSize(); j++)
			{
				Element = (*tagElement)[j];
				if(IS_ELEMENT_ID(KaxTagTargets))
				{
					KaxTagTargets *tagTargetsElement = (KaxTagTargets*)Element;					
					for (k = 0; k < tagTargetsElement->ListSize(); k++)
					{
						Element = (*tagTargetsElement)[k];
						if(IS_ELEMENT_ID(KaxTagTrackUID))
						{
							newTag.targetTrackUID = uint64(*static_cast<EbmlUInteger *>(Element));
							NOTE1("- TargetTrackUID : %I64d", newTag.targetTrackUID);
						}
						else if(IS_ELEMENT_ID(KaxTagEditionUID))
						{
							newTag.targetEditionUID = uint64(*static_cast<EbmlUInteger *>(Element));
							NOTE1("- TargetEditionUID : %I64d", newTag.targetEditionUID);
						}
						else if(IS_ELEMENT_ID(KaxTagChapterUID))
						{
							newTag.targetChapterUID = uint64(*static_cast<EbmlUInteger *>(Element));
							NOTE1("- TargetChapterUIDUID : %I64d", newTag.targetChapterUID);
						}
						else if(IS_ELEMENT_ID(KaxTagAttachmentUID))
						{
							newTag.targetAttachmentUID = uint64(*static_cast<EbmlUInteger *>(Element));
							NOTE1("- TargetAttachmentUID : %I64d", newTag.targetAttachmentUID);
						}
						else if(IS_ELEMENT_ID(KaxTagTargetTypeValue))
						{
							newTag.targetTypeValue = uint32(*static_cast<EbmlUInteger *>(Element));
							NOTE1("- TargetTypeValue : %d", newTag.targetTypeValue);
						}
						else if(IS_ELEMENT_ID(KaxTagTargetType))
						{
							newTag.targetType = std::string(*static_cast<KaxTagTargetType *>(Element));
							NOTE1("- TargetType : %s", newTag.targetType.c_str());
						}
					}
				}
				else if(IS_ELEMENT_ID(KaxTagSimple))
				{
					MatroskaSimpleTag newSimpleTag;
					KaxTagSimple *tagSimpleElement = (KaxTagSimple*)Element;
					NOTE("New SimpleTag");
					for (k = 0; k < tagSimpleElement->ListSize(); k++)
					{
						Element = (*tagSimpleElement)[k];
						if(IS_ELEMENT_ID(KaxTagName))
						{
							newSimpleTag.name = wcsupr((wchar_t *)UTFstring(*static_cast <EbmlUnicodeString *>(Element)).c_str());
							NOTE1("- Name : %s", newSimpleTag.name.GetUTF8().c_str());
						}
						else if(IS_ELEMENT_ID(KaxTagString))
						{
							newSimpleTag.value = UTFstring(*static_cast <EbmlUnicodeString *>(Element)).c_str();
							NOTE1("- Value : %s", newSimpleTag.value.GetUTF8().c_str());
						}
						else if(IS_ELEMENT_ID(KaxTagDefault))
						{
							newSimpleTag.defaultFlag = uint32(*static_cast<EbmlUInteger *>(Element));
							NOTE1("- TargetTypeValue : %d", newSimpleTag.default);
						}
						else if(IS_ELEMENT_ID(KaxTagLangue))
						{
							newSimpleTag.language = std::string(*static_cast <KaxTagLangue *>(Element));
							NOTE1("- Language : %s", newSimpleTag.language.c_str());
						}
						else if(IS_ELEMENT_ID(KaxTagSimple))
						{
							// ignore sub-tags
							// if we want "sub-tags" put this loop in another
							// function and call it recursively
						}
					}
					newTag.tags.push_back(newSimpleTag);
				}
			}
			m_Tags.push_back(newTag);
		}
	}
};

void MatroskaAudioParser::Parse_MetaSeek(EbmlElement *metaSeekElement, bool bInfoOnly) 
{
	uint64 lastSeekPos = 0;
	uint64 endSeekPos = 0;
	EbmlElement *l2 = NULL;
	EbmlElement *l3 = NULL;
	int UpperElementLevel = 0;

	if (metaSeekElement == NULL)
		return;

	l2 = m_InputStream.FindNextElement(metaSeekElement->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, true, 1);
	while (l2 != NULL) {
		if (UpperElementLevel > 0) {
			break;
		}
		if (UpperElementLevel < 0) {
			UpperElementLevel = 0;
		}

		if (EbmlId(*l2) == KaxSeek::ClassInfos.GlobalId) {
			//Wow we found the SeekEntries, time to speed up reading ;)
			l3 = m_InputStream.FindNextElement(l2->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, true, 1);

			EbmlId *id = NULL;
			while (l3 != NULL) {
				if (UpperElementLevel > 0) {
					break;
				}
				if (UpperElementLevel < 0) {
					UpperElementLevel = 0;
				}

				if (EbmlId(*l3) == KaxSeekID::ClassInfos.GlobalId) {
					binary *b = NULL;
					uint16 s = 0;
					KaxSeekID &seek_id = static_cast<KaxSeekID &>(*l3);
					seek_id.ReadData(m_InputStream.I_O(), SCOPE_ALL_DATA);
					b = seek_id.GetBuffer();
					s = (uint16)seek_id.GetSize();
					delete id;
					id = NULL;
					id = new EbmlId(b, s);

				} else if (EbmlId(*l3) == KaxSeekPosition::ClassInfos.GlobalId) {
					KaxSeekPosition &seek_pos = static_cast<KaxSeekPosition &>(*l3);
					seek_pos.ReadData(m_InputStream.I_O());				
					lastSeekPos = uint64(seek_pos);
					if (endSeekPos < lastSeekPos)
						endSeekPos = uint64(seek_pos);

					if ((*id == KaxCluster::ClassInfos.GlobalId) && !bInfoOnly) {
						//NOTE1("Found Cluster Seek Entry Postion: %u", (unsigned long)lastSeekPos);
						//uint64 orig_pos = inputFile.getFilePointer();									
						MatroskaMetaSeekClusterEntry newCluster;
						newCluster.timecode = MAX_UINT64;
						newCluster.filePos = static_cast<KaxSegment *>(m_ElementLevel0)->GetGlobalPosition(lastSeekPos);
						m_ClusterIndex.push_back(newCluster);

					} else if ((*id == KaxSeekHead::ClassInfos.GlobalId) && !bInfoOnly) {
						NOTE1("Found MetaSeek Seek Entry Postion: %u", (unsigned long)lastSeekPos);
						uint64 orig_pos = m_IOCallback.getFilePointer();									
						m_IOCallback.setFilePointer(static_cast<KaxSegment *>(m_ElementLevel0)->GetGlobalPosition(lastSeekPos));
						
						EbmlElement *levelUnknown = m_InputStream.FindNextID(KaxSeekHead::ClassInfos, 0xFFFFFFFFFFFFFFFFL);										
						Parse_MetaSeek(levelUnknown, false);

						m_IOCallback.setFilePointer(orig_pos);
					}

				} else {

				}
				l3->SkipData(m_InputStream, l3->Generic().Context);
				delete l3;
				l3 = NULL;
				l3 = m_InputStream.FindNextElement(l2->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, true, 1);
			} // while (l3 != NULL)
			delete id;
			id = NULL;
		} else {

		}

		if (UpperElementLevel > 0) {    // we're coming from l3
			UpperElementLevel--;
			delete l2;
			l2 = l3;
			if (UpperElementLevel > 0)
				break;

		} else {
			l2->SkipData(m_InputStream, l2->Generic().Context);
			delete l2;
			l2 = NULL;
			l2 = m_InputStream.FindNextElement(metaSeekElement->Generic().Context, UpperElementLevel, 0xFFFFFFFFFFFFFFFFL, true, 1);
		}
	} // while (l2 != NULL)
}

uint64 MatroskaAudioParser::GetClusterTimecode(uint64 filePos) {	
	try {
		uint64 ret = MAX_UINT64;

		int UpperElementLevel = 0;
		// Elements for different levels
		EbmlElement *ElementLevel1 = NULL;
		EbmlElement *ElementLevel2 = NULL;		
		EbmlElement *ElementLevel3 = NULL;		

		m_IOCallback.setFilePointer(filePos);
		// Find the element data
		ElementLevel1 = m_InputStream.FindNextID(KaxCluster::ClassInfos, 0xFFFFFFFFFFFFFFFFL);
		if (ElementLevel1 == NULL)
			return MAX_UINT64;

		if (EbmlId(*ElementLevel1) == KaxCluster::ClassInfos.GlobalId) {
			//KaxCluster *SegmentCluster = static_cast<KaxCluster *>(ElementLevel1);
			//uint32 ClusterTimecode = 0;

			// read blocks and discard the ones we don't care about
			ElementLevel2 = m_InputStream.FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->ElementSize(), false);
			while (ElementLevel2 != NULL) {
				if (UpperElementLevel > 0) {
					break;
				}
				if (UpperElementLevel < 0) {
					UpperElementLevel = 0;
				}
				if (EbmlId(*ElementLevel2) == KaxClusterTimecode::ClassInfos.GlobalId) {						
					KaxClusterTimecode & ClusterTime = *static_cast<KaxClusterTimecode*>(ElementLevel2);
					ClusterTime.ReadData(m_InputStream.I_O());
					ret = uint64(ClusterTime) * m_TimecodeScale;
					
				}

				if (UpperElementLevel > 0) {
					UpperElementLevel--;
					delete ElementLevel2;
					ElementLevel2 = ElementLevel3;
					if (UpperElementLevel > 0)
						break;
				} else {
					ElementLevel2->SkipData(m_InputStream, ElementLevel2->Generic().Context);
					//if (ElementLevel2 != pChecksum)
					delete ElementLevel2;								
					ElementLevel2 = NULL;
					if (ret == -1)
						ElementLevel2 = m_InputStream.FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->ElementSize(), false);
				}
			}
		}
		delete ElementLevel1;

		return ret;
	} catch (...) {
		return MAX_UINT64;
	}	
};

MatroskaMetaSeekClusterEntry *MatroskaAudioParser::FindCluster(uint64 timecode)
{
	try {
		#ifdef _DEBUG_NO_SEEKING
		static size_t callCount = 0;
		if (callCount < m_ClusterIndex.size())
			return &m_ClusterIndex[callCount++];
		else return NULL;
		#endif

		if (timecode == 0)
			// Special case
			return &m_ClusterIndex.at(0);
		
		MatroskaMetaSeekClusterEntry *correctEntry = NULL;
		double clusterDuration = (double)(int64)m_ClusterIndex.size() / (double)(int64)m_Duration;
		size_t clusterIndex = clusterDuration * (double)(int64)timecode;
		//int lookCount = 0;
		if (clusterIndex > m_ClusterIndex.size())
			clusterIndex = m_ClusterIndex.size()-1;

		MatroskaMetaSeekClusterEntry *clusterEntry = NULL;
		MatroskaMetaSeekClusterEntry *prevClusterEntry = NULL;
		MatroskaMetaSeekClusterEntry *nextClusterEntry = NULL;
		while (correctEntry == NULL) {
			clusterEntry = &m_ClusterIndex.at(clusterIndex);		
			if (clusterIndex > 0)
				prevClusterEntry = &m_ClusterIndex.at(clusterIndex-1);
			if (clusterIndex+1 < m_ClusterIndex.size())
				nextClusterEntry = &m_ClusterIndex.at(clusterIndex+1);			
			
			// We need timecodes to do good seeking
			if (clusterEntry->timecode == MAX_UINT64) {				
				clusterEntry->timecode = GetClusterTimecode(clusterEntry->filePos);			
			}
			if (prevClusterEntry != NULL && prevClusterEntry->timecode == MAX_UINT64) {
				prevClusterEntry->timecode = GetClusterTimecode(prevClusterEntry->filePos);			
			}
			if (nextClusterEntry != NULL && nextClusterEntry->timecode == MAX_UINT64) {				
				nextClusterEntry->timecode = GetClusterTimecode(nextClusterEntry->filePos);			
			}

			if (clusterEntry->timecode == timecode) {
				// WOW, we are seeking directly to this cluster
				correctEntry = clusterEntry;
				break;
			}

			if (prevClusterEntry != NULL) {
				if (clusterEntry->timecode > timecode && timecode > prevClusterEntry->timecode) {
					// We found it !!!
					correctEntry = prevClusterEntry;
					break;
				}
				if (prevClusterEntry->timecode == timecode) {
					// WOW, we are seeking directly to this cluster
					correctEntry = prevClusterEntry;
					break;
				}
				// Check if we overshot the needed cluster
				if (timecode < prevClusterEntry->timecode) {
					clusterIndex--;
					//lookCount++; // This is how many times we have 'looked'
					continue;
				}
			}
			
			if (nextClusterEntry != NULL) {
				if (clusterEntry->timecode < timecode && timecode < nextClusterEntry->timecode) {
					// We found it !!!
					correctEntry = clusterEntry;
					break;
				}			
				if (nextClusterEntry->timecode == timecode) {
					// WOW, we are seeking directly to this cluster
					correctEntry = nextClusterEntry;
					break;
				}
				// Check if we undershot the needed cluster
				if (timecode > nextClusterEntry->timecode) {
					clusterIndex++;
					//lookCount--;
					continue;
				}
			}
			// We should never get here, unless this is the last cluster
			assert(clusterEntry != NULL);	
			if (timecode <= m_Duration)
				correctEntry = clusterEntry;
			else
				break;
		}

		if (correctEntry != NULL)
			NOTE3("MatroskaAudioParser::FindCluster(timecode = %u) seeking to cluster %i at %u", (uint32)(timecode / m_TimecodeScale), (uint32)correctEntry->clusterNo, (uint32)correctEntry->filePos);
		else
			NOTE1("MatroskaAudioParser::FindCluster(timecode = %u) seeking failed", (uint32)(timecode / m_TimecodeScale));

		return correctEntry;
	} catch (...) {
		return NULL;
	}
}

void MatroskaAudioParser::CountClusters() 
{
	for (uint32 c = 0; c < m_ClusterIndex.size(); c++) {
		MatroskaMetaSeekClusterEntry &clusterEntry = m_ClusterIndex.at(c);		
		clusterEntry.clusterNo = c;
	}
}

void MatroskaAudioParser::FixChapterEndTimes()
{
	if (m_Chapters.size() > 0) {
		MatroskaChapterInfo *nextChapter = &m_Chapters.at(m_Chapters.size()-1);
		if (nextChapter->timeEnd == 0) {
			nextChapter->timeEnd = m_Duration;
		}
		for (uint32 c = 0; c < m_Chapters.size()-1; c++) {
			MatroskaChapterInfo &currentChapter = m_Chapters.at(c);	
			nextChapter = &m_Chapters.at(c+1);
			if (currentChapter.timeEnd == 0) {
				currentChapter.timeEnd = nextChapter->timeStart;
			}
		}
		nextChapter = &m_Chapters.at(m_Chapters.size()-1);
		if ((nextChapter->timeEnd == 0) || (nextChapter->timeEnd == nextChapter->timeStart)) {
			nextChapter->timeEnd = m_Duration;
		}
	}	
}

bool MatroskaAudioParser::FindEditionUID(uint64 uid)
{
	for (uint32 c = 0; c < m_Editions.size(); c++) {
		MatroskaEditionInfo &currentEdition = m_Editions.at(c);	
		if (currentEdition.editionUID == uid)
			return true;
	}	
	return false;
}

bool MatroskaAudioParser::FindChapterUID(uint64 uid)
{
	for (uint32 c = 0; c < m_Chapters.size(); c++) {
		MatroskaChapterInfo &currentChapter = m_Chapters.at(c);	
		if (currentChapter.chapterUID == uid)
			return true;
	}	
	return false;
}

bool MatroskaSearch::Skip()
{
	int j;

	for (j = 0; j < SEARCH_TABLE_SIZE; j++) skip[j] = SEARCH_PATTERN_SIZE;
	for (j = 0; j < SEARCH_PATTERN_SIZE - 1; j++)
		skip[pattern[j] & 0x00ff] = SEARCH_PATTERN_SIZE-1-j;
	return true;
}

bool MatroskaSearch::Next()
{
	int  j, k, s;
	int  *g;

	if ((g = (int *)malloc(sizeof(int)*SEARCH_PATTERN_SIZE)) == NULL) return false;
	for (j = 0; j < SEARCH_PATTERN_SIZE; j++) next[j] = 2*SEARCH_PATTERN_SIZE - 1 - j;
	j = SEARCH_PATTERN_SIZE;
	for (k = SEARCH_PATTERN_SIZE - 1; k >= 0; k--) {
		g[k] = j;
		while (j != SEARCH_PATTERN_SIZE && pattern[j] != pattern[k]) {
			next[j] = (next[j] <= SEARCH_PATTERN_SIZE-1-k) ? next[j] : SEARCH_PATTERN_SIZE-1-k;
			j = g[j];
		}
		j--;
	}
	s = j;
	for (j = 0; j < SEARCH_PATTERN_SIZE; j++) {
		next[j] = (next[j] <= s+SEARCH_PATTERN_SIZE-j) ? next[j] : s+SEARCH_PATTERN_SIZE-j;
		if (j >= s) s = g[s];
	}
	free(g);
	return true;
}

int MatroskaSearch::Match(unsigned int start)
{
	int i, j;

	i = SEARCH_PATTERN_SIZE - 1 + start;
	while (i < SEARCH_SOURCE_SIZE) {
		j = SEARCH_PATTERN_SIZE - 1;
		while (j >= 0 && source[i] == pattern[j]) {
			i--;
			j--;
		}
		if (j < 0) return i + 1;
		if (skip[source[i] & 0x00ff] >= next[j])
			i += skip[source[i] & 0x00ff];
		else i += next[j];
	}
	return -1;
};

}