/*
 *  Part of the foobar2000 Matroska plugin
 *
 *  Copyright (C) Jory Stone (jcsston at toughguy net) - 2003-2004
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 *  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*!
    \file matroska_parser.h
		\version $Id$
    \brief An audio slated Matroska Reader+Parser
		\author Jory Stone     <jcsston @ toughguy.net>
*/

#ifndef _MATROSKA_PARSER_H_
#define _MATROSKA_PARSER_H_

#define MULTITRACK 1

#include "../SDK/foobar2000.h"
#include "../../pfc/pfc.h"
#include "Foobar2000ReaderIOCallback.h"
//#include "DbgOut.h"
#include <queue>
#include <deque>

// libebml includes
#include "ebml/StdIOCallback.h"
#include "ebml/EbmlTypes.h"
#include "ebml/EbmlHead.h"
#include "ebml/EbmlVoid.h"
#include "ebml/EbmlCrc32.h"
#include "ebml/EbmlSubHead.h"
#include "ebml/EbmlStream.h"
#include "ebml/EbmlBinary.h"
#include "ebml/EbmlString.h"
#include "ebml/EbmlUnicodeString.h"
#include "ebml/EbmlContexts.h"
#include "ebml/EbmlVersion.h"

// libmatroska includes
#include "matroska/KaxConfig.h"
#include "matroska/KaxBlock.h"
#include "matroska/KaxSegment.h"
#include "matroska/KaxContexts.h"
#include "matroska/KaxSeekHead.h"
#include "matroska/KaxTracks.h"
#include "matroska/KaxInfo.h"
#include "matroska/KaxInfoData.h"
#include "matroska/KaxTags.h"
#include "matroska/KaxTag.h"
#include "matroska/KaxTagMulti.h"
#include "matroska/KaxCluster.h"
#include "matroska/KaxClusterData.h"
#include "matroska/KaxTrackAudio.h"
#include "matroska/KaxTrackVideo.h"
#include "matroska/KaxAttachments.h"
#include "matroska/KaxAttached.h"
#include "matroska/KaxChapters.h"
#include "matroska/KaxVersion.h"

#define TIMECODE_SCALE  1000000
#define MAX_UINT64 0xFFFFFFFFFFFFFFFF

//Memory Leak Debuging define
#ifdef _DEBUG
#define _CRTDBG_MAP_ALLOC 
#include <stdlib.h>
#include <crtdbg.h>
#endif

#ifdef _DEBUG
//   #define DEBUG_CLIENTBLOCK   new( _CLIENT_BLOCK, __FILE__, __LINE__)
#else
   #define DEBUG_CLIENTBLOCK
#endif // _DEBUG

#ifdef _DEBUG
//#define new DEBUG_CLIENTBLOCK
#include <boost/timer.hpp>
#include <tchar.h>
#include <stdarg.h>
#define HPRINTF_MES_LENGTH 256
inline void hprintf( const _TCHAR * format, ... )
{
    _TCHAR buff[HPRINTF_MES_LENGTH];
    va_list vl;
    va_start(vl, format);
    _vsntprintf(buff, HPRINTF_MES_LENGTH, format, vl);
    va_end(vl);
    OutputDebugString(buff);
}
#define TIMER boost::timer t
#define RETIMER t.restart()
#define _TIMER(x) hprintf(L"%S: %f sec.\n",x,t.elapsed())
#else
#define hprintf __noop
#define TIMER __noop
#define RETIMER __noop
#define _TIMER(x) __noop
#endif

#define NOTE(x)
#define NOTE1(x, y1)
#define NOTE2(x, y1, y2)
#define NOTE3(x, y1, y2, y3)
#define NOTE4(x, y1, y2, y3, y4)

using namespace LIBEBML_NAMESPACE;
using namespace LIBMATROSKA_NAMESPACE;

typedef std::vector<uint8> ByteArray;

namespace Matroska {

class MatroskaAttachment {
public:
	MatroskaAttachment();

	UTFstring FileName;
	std::string MimeType;
	UTFstring Description;
	UTFstring SourceFilename;
	uint64 SourceStartPos;
	uint64 SourceDataLength;
};

class MatroskaSimpleTag {
public:
	MatroskaSimpleTag();

	UTFstring name;
	UTFstring value;
	uint32 defaultFlag;
	std::string language;

	bool hidden;
	bool removalPending;
};

class MatroskaTagInfo {
public:
	MatroskaTagInfo();
	void SetTagValue(const char *name, const char *value, int index = 0);
	void MarkAllAsRemovalPending();
	void RemoveMarkedTags();
	
	uint64 targetTrackUID;
	uint64 targetEditionUID;
	uint64 targetChapterUID;
	uint64 targetAttachmentUID;
	uint32 targetTypeValue;
	std::string targetType;

	std::vector<MatroskaSimpleTag> tags;
};

class MatroskaEditionInfo {
public:
	MatroskaEditionInfo();

	uint64 editionUID;
	/// Vector of all the tracks this edition applies to
	/// if it's empty then this edition applies to all tracks
	std::vector<uint64> tracks;
};

struct MatroskaChapterDisplayInfo {
	MatroskaChapterDisplayInfo();

	UTFstring string;
	std::string lang;
	std::string country;
};

class MatroskaChapterInfo {
public:
	MatroskaChapterInfo();

	uint64 chapterUID;
	uint64 timeStart;
	uint64 timeEnd;
	/// Vector of all the tracks this chapter applies to
	/// if it's empty then this chapter applies to all tracks
	std::vector<uint64> tracks;
	/// Vector of strings we can display for chapter
	std::vector<MatroskaChapterDisplayInfo> display;
};

struct MatroskaMetaSeekClusterEntry {
	uint32 clusterNo;
	uint64 filePos;
	uint64 timecode;
};

class MatroskaAudioParser {
public:
	MatroskaAudioParser(service_ptr_t<file> input, abort_callback & p_abort);
	~MatroskaAudioParser();

	/// The main header parsing function
	/// \return 0 File parsed ok
	/// \return 1 Failed
	int Parse(bool bInfoOnly = false, bool bBreakAtClusters = true);

	UTFstring GetSegmentFileName() { return m_SegmentFilename; }
	std::vector<MatroskaAttachment> &GetAttachmentList() { return m_AttachmentList; }

protected:
	void Parse_MetaSeek(EbmlElement *metaSeekElement, bool bInfoOnly);
	void Parse_Chapters(KaxChapters *chaptersElement);
	void Parse_Chapter_Atom(KaxChapterAtom *ChapterAtom);
	void Parse_Tags(KaxTags *tagsElement);
	int FillQueue();
	uint64 GetClusterTimecode(uint64 filePos);
	MatroskaMetaSeekClusterEntry *FindCluster(uint64 timecode);
	void CountClusters();
	void FixChapterEndTimes();
	// See if the edition uid is already in our vector
	// \return true Yes, we already have this uid
	// \return false Nope
	bool FindEditionUID(uint64 uid);
	// See if the chapter uid is already in our vector
	// \return true Yes, we already have this uid
	// \return false Nope
	bool FindChapterUID(uint64 uid);
	/// Adds the info tags to the current file in memory 
	void AddTags(file_info &info);
	/// Set the info tags to the current tags file in memory
	void SetTags(const file_info &info);
	bool AreTagsIdenticalAtAllLevels(const char * name);
	bool AreTagsIdenticalAtEditionLevel(const char * name);
	bool AreTagsIdenticalAtChapterLevel(const char * name);
	void MarkHiddenTags();		

	Foobar2000ReaderIOCallback m_IOCallback;
	EbmlStream m_InputStream;
	/// The main/base/master element, should be the segment
	EbmlElement *m_ElementLevel0;

	std::vector<MatroskaEditionInfo> m_Editions;
	std::vector<MatroskaChapterInfo> m_Chapters;
	std::vector<MatroskaTagInfo> m_Tags;

	/// This is the index of clusters in the file, it's used to seek in the file
	std::vector<MatroskaMetaSeekClusterEntry> m_ClusterIndex;

	uint32 m_CurrentTrackNo;
	std::vector<MatroskaAttachment> m_AttachmentList;

	uint64 m_CurrentTimecode;
	uint64 m_Duration;
	uint64 m_TimecodeScale;
	UTFstring m_WritingApp;
	UTFstring m_MuxingApp;
	UTFstring m_FileTitle;
	int32 m_FileDate;
	UTFstring m_SegmentFilename;

	uint64 m_FileSize;
	uint64 m_TagPos;
	uint32 m_TagSize;
	uint32 m_TagScanRange;

	//mem_block_factalloc<BYTE> m_framebuffer;
	//int UpperElementLevel;
private:
	// We have no need for these
	//MatroskaAudioParser(const MatroskaAudioParser &refParser) { };
	//MatroskaAudioParser &operator=(const MatroskaAudioParser &refParser) { return *this; };
};

class MatroskaSearch
{
private:
	static const int SEARCH_SOURCE_SIZE = 1024*64;
	static const int SEARCH_TABLE_SIZE = SEARCH_SOURCE_SIZE;
	static const int SEARCH_PATTERN_SIZE = 3;
    binary * source, * pattern;
	int pos;
    int  next[SEARCH_TABLE_SIZE], skip[SEARCH_TABLE_SIZE];
	bool Skip();
	bool Next();
public:
	MatroskaSearch(binary * p_source, binary * p_pattern)
	{
		ZeroMemory(next, sizeof(next));
		ZeroMemory(skip, sizeof(skip));
		source = p_source;
		pattern = p_pattern;
		Skip();
		Next();
	}
	~MatroskaSearch()
	{
		delete[] source;
		delete[] pattern;
	};
	int Match(unsigned int start = 0);
};

}

#endif // _MATROSKA_PARSER_H_
